"""Log handler and formatter to log messages to MongoDB

Not (yet) guaranteed to work with arbitrary extras fields"""

import uuid
import traceback
import logging

from datetime import datetime
from socket import gethostname

class JsonFormatter(logging.Formatter):
    def __init__(self, **kwargs):
        self.host = gethostname()
        self.extra = kwargs

    def format(self, record):
        # We want a dict and not a LogRecord object.
        data = record.__dict__.copy()

        if record.args:
            msg = record.msg % record.args
        else:
            msg = record.msg

        args = record.args if isinstance(record.args, tuple) else (record.args,)

        data.update(
            time=datetime.utcnow(),
            host=self.host,
            message=msg,
            args=tuple(unicode(arg) for arg in args))
        data.update(self.extra)
        if data.get('exc_info'):
            data['exc_info'] = self.formatException(data['exc_info'])

        return data


class MongoHandler(logging.Handler):
    """Meant to be used with the MongoFormatter, but any formatter that gives
    records that MongoDB will accept should work.
    Mandatory first argument is a collection object from MongoDB"""
    def __init__(self, collection, formatter=None, level=logging.NOTSET):
        logging.Handler.__init__(self, level)
        self.collection = collection
        if formatter == None:
            formatter = JsonFormatter()
        self.formatter = formatter

    def emit(self, record):
        self.collection.insert(self.format(record))



"""Tools to make some minimal guarantee of homogenous
data in a MongoDB collection"""

from __future__ import print_function
import pymongo
import sys

def typecheck_collection(collection, fields):
    """Assert matching types"""
    def checker(document, fields):
        for name, type_ in fields.items():
            if name in document:
                if not isinstance(document[name], type_):
                    raise TypeError('Expected {}, got {}'.format(
                        type_, type(document[name])))
    return wrap_collection(collection, checker, fields)

def convert_collection(collection, fields):
    """Ensure types by converting them"""
    def checker(document, fields):
        for name, type_ in fields.items():
            if name in document:
                document[name] = type_(document[name])
    return wrap_collection(collection, checker, fields)

def limit_collection(collection, fields):
    """Ensure only given fields are present"""
    def checker(document, fields):
        for name in document:
            if name not in fields:
                raise TypeError(
                    'Found unexpected key {}'.format(name))
    return wrap_collection(collection, checker, fields)

def checked_collection(collection, fields):
    """Ensure given fields are present"""
    def checker(document, fields):
        for name in fields:
            if name not in document:
                raise TypeError(
                    'Missing required key {}'.format(name))

    return wrap_collection(collection, checker, fields)


def wrap_collection(collection, checker, fields):
    # Don't wrap at all if optimize is on
    if sys.flags.optimize > 0:
        return collection

    updater = collection.update
    inserter = collection.insert

    def update(spec, document, **kwargs):
        checker(document, fields)
        return updater(spec, document, **kwargs)

    def insert(document, *args, **kwargs):
        checker(document, fields)
        return inserter(document, *args, **kwargs)

    collection.update = update
    collection.insert = insert

    return collection



"""Locking using MongoDB

Not meant to be extremely fast, but a convenient way to get distributed
locks if you have MongoDB in your stack"""
import contextlib
import datetime
import socket

HOSTNAME = socket.gethostname()


class LockException(Exception):
    def __init__(self, message, lock_document=None):
        super(LockException, self).__init__(message)
        self.lock_document = lock_document

class Locked(LockException):
    pass

class ConcurrentLockModificationError(LockException):
    pass


def engage_lock(lock_collection, lock_id, lock_subdoc):
    """Direct access to locking mechanism, you probably should stick with
    Lock.lock or the context manager 'lock' instead"""
    document = lock_collection.find_one({'_id': lock_id})
    if not document:
        try:
            lock_collection.insert(
                {'_id': lock_id,
                 lock_subdoc: {'locked': False,
                             'updated': datetime.datetime.utcnow(),
                             'version': 1}})
            document = lock_collection.find_one({'_id': lock_id})
        except KeyError:
            pass  # Concurrent creation
    if not document.get(lock_subdoc):
        try:
            lock_collection.update(
                {'_id': lock_id},
                {lock_subdoc: {'locked': False,
                             'updated': datetime.datetime.utcnow(),
                             'version': 1}})
            document = lock_collection.find_one({'_id': lock_id})
        except KeyError:
            pass

    if document.get(lock_subdoc).get('locked'):
        raise Locked('Lock is already locked', document)

    write_result = lock_collection.update({
        '_id': lock_id,
        lock_subdoc + '.version': document[lock_subdoc]['version'],
        lock_subdoc + '.locked': document[lock_subdoc]['locked']
    }, {
        '$set': {lock_subdoc + '.locked': True,
                 lock_subdoc + '.updated': datetime.datetime.utcnow(),
                 lock_subdoc + '.host': HOSTNAME},
        '$inc': {lock_subdoc + '.version': 1}
    })

    if write_result.get('n') != 1:
        raise Locked(lock_collection.find_one({'_id': lock_id}))
    return document

def release_lock(lock_collection, lock_id, lock_subdoc, document):
    """Direct access to unlocking mechanism, you probably should stick with
    Lock.unlock or the context manager 'lock' instead"""

    write_result = lock_collection.update(
        {
            '_id': lock_id,
            lock_subdoc + '.version': document[lock_subdoc]['version'] + 1,
            lock_subdoc + '.locked': True
        },
        {
            '$set': {lock_subdoc + '.locked': False,
                     lock_subdoc + '.updated': datetime.datetime.utcnow()},
            '$inc': {lock_subdoc + '.version': 1}
        }
    )
    if write_result.get('n') != 1:
        raise ConcurrentLockModificationError(
            lock_collection.find_one({'_id': lock_id}))


class Lock(object):
    """Lock as an object"""
    def __init__(self, lock_collection, lock_id, lock_subdoc=None):
        self._lock_collection = lock_collection
        self._lock_id = lock_id
        self._lock_subdoc = lock_subdoc
        self._document = None
        self.is_locked = False  # Informational only!

    def lock(self):
        self._document = engage_lock(self._lock_collection,
                                     self._lock_id, self._lock_subdoc)
        self.is_locked = True

    def unlock(self):
        if self._document is None:
            raise LockException('Not locked')
        release_lock(self._lock_collection, self._lock_id, self._lock_subdoc,
                self._document)
        self.is_locked = False

    def __repr__(self):
        return 'Lock {}[{}].{} status {}'.format(self._lock_collection,
                                                 self._lock_id,
                                                 self._lock_subdoc,
                                                 self.is_locked)

@contextlib.contextmanager
def lock(lock_collection, lock_id, lock_subdoc):
    """Lock as a context manager"""
    engage_lock(lock_collection, lock_id, lock_subdoc)
    try:
        yield  # This is where the calling code runs
    finally:
        release_lock(lock_collection, lock_id, lock_subdoc)



from setuptools import setup
import shutil, os
VERSION = '1.0.1'
SDIST = 'ma.mongo-' + VERSION

os.mkdir(SDIST)
shutil.copyfile('setup.mongo.py', SDIST +'/setup.py')

setup(
    name='ma.mongo',
    version = VERSION,
    description='MongoDB utilities',
    author='Fredrik Haard',
    author_email='fredrik@metallapan.se',
    packages=['ma.mongo']
)
